/*
(C) 2002 Advanced Micro Devices, Inc. 
** YOUR USE OF THIS LIBRARY IS SUBJECT TO THE TERMS 
    AND CONDITIONS OF THE GNU LESSER GENERAL PUBLIC 
    LICENSE FOUND IN THE "README" FILE THAT IS INCLUDED WITH 
    THIS LIBRARY**
*/


#include "libm_amd.h"
#include "libm_util_amd.h"

#define USE_REMAINDER_PIBY2F_INLINE
#define USE_VAL_WITH_FLAGS
#define USE_NAN_WITH_FLAGS
#include "libm_inlines_amd.h"
#undef USE_VAL_WITH_FLAGS
#undef USE_NAN_WITH_FLAGS
#undef USE_REMAINDER_PIBY2F_INLINE

/* sin(x) approximation valid on the interval [-pi/4,pi/4]. */
static inline double sinf_piby4(double x)
{
  /* Taylor series for sin(x) is x - x^3/3! + x^5/5! - x^7/7! ...
                          = x * (1 - x^2/3! + x^4/5! - x^6/7! ...
                          = x * f(w)
     where w = x*x and f(w) = (1 - w/3! + w^2/5! - w^3/7! ...
     We use a minimax approximation of (f(w) - 1) / w
     because this produces an expansion in even powers of x.
  */
  double x2;
  static const double
  c1 = -0.166666666638608441788607926e0,
  c2 = 0.833333187633086262120839299e-2,
  c3 = -0.198400874359527693921333720e-3,
  c4 = 0.272500015145584081596826911e-5;

  x2 = x * x;
  return (x + x * x2 * (c1 + x2 * (c2 + x2 * (c3 + x2 * c4))));
}

/* cos(x) approximation valid on the interval [-pi/4,pi/4]. */
static inline double cosf_piby4(double x)
{
  /* Taylor series for cos(x) is 1 - x^2/2! + x^4/4! - x^6/6! ...
                          = f(w)
     where w = x*x and f(w) = (1 - w/2! + w^2/4! - w^3/6! ...
     We use a minimax approximation of (f(w) - 1 + w/2) / (w*w)
     because this produces an expansion in even powers of x.
  */
  double x2;
  static const double
    c1 = 0.41666666664325175238031e-1,
    c2 = -0.13888887673175665567647e-2,
    c3 = 0.24800600878112441958053e-4,
    c4 = -0.27301013343179832472841e-6;

  x2 = x * x;
  return (1.0 - 0.5 * x2 + (x2 * x2 *
                      (c1 + x2 * (c2 + x2 * (c3 + x2 * c4)))));
}


void __sincosf(float x, float *s, float *c)
{
  double r, dx;
  int region, xneg;

  unsigned long ux, ax;

  dx = x;

  GET_BITS_DP64(dx, ux);
  ax = (ux & ~SIGNBIT_DP64);

  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f80000000000000) /* abs(x) < 2.0^(-7) */
        {
          if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
            {
              if (ax == 0x0000000000000000)
                {
                  *s = dx;
                  *c = 1.0;
                }
              else
                {
                  *s = val_with_flags(dx, AMD_F_INEXACT);
                  *c = val_with_flags(1.0, AMD_F_INEXACT);
                }
            }
          else
            {
              *s = dx - dx*dx*dx*0.166666666666666666;
              *c = 1.0 - dx*dx*0.5;
            }
        }
      else
        {
          *s = sinf_piby4(x);
          *c = cosf_piby4(x);
        }
      return;
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        *s = *c = dx + dx; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        *s = *c = nan_with_flags(AMD_F_INVALID);
      return;
    }

  xneg = (ux >> 63);

  if (xneg)
    dx = -dx;

  /* Reduce abs(x) into range [-pi/4,pi/4] */
  __remainder_piby2f_inline(dx, ax, &r, &region);

  if (xneg)
    {
      switch (region)
        {
        default:
        case 0:
          *s = -sinf_piby4(r);
          *c = cosf_piby4(r);
          break;
        case 1:
          *s = -cosf_piby4(r);
          *c = -sinf_piby4(r);
          break;
        case 2:
          *s = sinf_piby4(r);
          *c = -cosf_piby4(r);
          break;
        case 3:
          *s = cosf_piby4(r);
          *c = sinf_piby4(r);
          break;
        }
    }
  else
    {
      switch (region)
        {
        default:
        case 0:
          *s = sinf_piby4(r);
          *c = cosf_piby4(r);
          break;
        case 1:
          *s = cosf_piby4(r);
          *c = -sinf_piby4(r);
          break;
        case 2:
          *s = -sinf_piby4(r);
          *c = -cosf_piby4(r);
          break;
        case 3:
          *s = -cosf_piby4(r);
          *c = sinf_piby4(r);
          break;
        }
    }
}

float __sinf(float x)
{
  double r, dx;
  int region, xneg;

  unsigned long ux, ax;

  dx = x;

  GET_BITS_DP64(dx, ux);
  ax = (ux & ~SIGNBIT_DP64);

  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f80000000000000) /* abs(x) < 2.0^(-7) */
        {
          if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
            {
              if (ax == 0x0000000000000000)
                return x;
              else
                return val_with_flags(dx, AMD_F_INEXACT);
            }
          else
            return x - x*x*x*0.166666666666666666;
        }
      else
        return sinf_piby4(dx);
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        return x + x; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        return nan_with_flags(AMD_F_INVALID);
    }

  xneg = (ux >> 63);

  if (xneg)
    dx = -dx;

  /* Reduce abs(x) into range [-pi/4,pi/4] */
  __remainder_piby2f_inline(dx, ax, &r, &region);

  if (xneg)
    {
      switch (region)
        {
        default:
        case 0: return -sinf_piby4(r);
        case 1: return -cosf_piby4(r);
        case 2: return sinf_piby4(r);
        case 3: return cosf_piby4(r);
        }
    }
  else
    {
      switch (region)
        {
        default:
        case 0: return sinf_piby4(r);
        case 1: return cosf_piby4(r);
        case 2: return -sinf_piby4(r);
        case 3: return -cosf_piby4(r);
        }
    }
}

#if 1
/* Stupidly, computing cosf via sincosf is much faster,
   even though sincosf does the same work and more. */
float __cosf(float x)
{
  float s, c;
  __sincosf(x, &s, &c);
  return c;
}

#else
/* This is the way cosf should be done, but it runs half
   as fast as it ought to */

float __cosf(float x)
{
  double r, dx;
  int region, xneg;

  unsigned long ux, ax;

  dx = x;

  GET_BITS_DP64(dx, ux);
  ax = (ux & ~SIGNBIT_DP64);

  if (ax <= 0x3fe921fb54442d18) /* abs(x) <= pi/4 */
    {
      if (ax < 0x3f80000000000000) /* abs(x) < 2.0^(-7) */
        {
          if (ax < 0x3f20000000000000) /* abs(x) < 2.0^(-13) */
            {
              if (ax == 0x0000000000000000)
                return 1.0F;
              else
                return val_with_flags(1.0, AMD_F_INEXACT);
            }
          else
            return 1.0F - x*x*0.5F;
        }
      else
        return cosf_piby4(dx);
    }
  else if ((ux & EXPBITS_DP64) == EXPBITS_DP64)
    {
      /* x is either NaN or infinity */
      if (ux & MANTBITS_DP64)
        /* x is NaN */
        return x + x; /* Raise invalid if it is a signalling NaN */
      else
        /* x is infinity. Return a NaN */
        return nan_with_flags(AMD_F_INVALID);
    }

  xneg = (ux >> 63);

  if (xneg)
    dx = -dx;

  /* Reduce abs(x) into range [-pi/4,pi/4] */
  __remainder_piby2f_inline(dx, ax, &r, &region);

  switch (region)
    {
    default:
    case 0: return cosf_piby4(r);
    case 1: return -sinf_piby4(r);
    case 2: return -cosf_piby4(r);
    case 3: return sinf_piby4(r);
    }
}
#endif

weak_alias (__sinf, sinf)
weak_alias (__cosf, cosf)
weak_alias (__sincosf, sincosf)
