/*
(C) 2002 Advanced Micro Devices, Inc. 
** YOUR USE OF THIS LIBRARY IS SUBJECT TO THE TERMS 
    AND CONDITIONS OF THE GNU LESSER GENERAL PUBLIC 
    LICENSE FOUND IN THE "README" FILE THAT IS INCLUDED WITH 
    THIS LIBRARY**
*/

#include "libm_amd.h"
#include "libm_util_amd.h"

#define USE_ZEROF_WITH_FLAGS
#define USE_INFINITYF_WITH_FLAGS
#define USE_NANF_WITH_FLAGS
#include "libm_inlines_amd.h"
#undef USE_ZEROF_WITH_FLAGS
#undef USE_INFINITYF_WITH_FLAGS
#undef USE_NANF_WITH_FLAGS

/* Deal with errno for out-of-range result */
#include "libm_errno_amd.h"
static inline float retval_errno_erange_overflow(float x, float y, int sign)
{
  struct exception exc;
  exc.arg1 = (double)x;
  exc.arg2 = (double)y;
  exc.type = OVERFLOW;
  exc.name = (char *)"powf";
  if (_LIB_VERSION == _SVID_)
    {
      if (sign == 1)
        exc.retval = HUGE;
      else /* sign = -1 */
        exc.retval = -HUGE;
    }
  else
    {
      if (sign == 1)
        exc.retval = infinityf_with_flags(AMD_F_OVERFLOW);
      else /* sign == -1 */
        exc.retval = -infinityf_with_flags(AMD_F_OVERFLOW);
    }
  if (_LIB_VERSION == _POSIX_)
    __set_errno(ERANGE);
  else if (!matherr(&exc))
    __set_errno(ERANGE);
  return exc.retval;
}

static inline float retval_errno_erange_underflow(float x, float y, int sign)
{
  struct exception exc;
  exc.arg1 = (double)x;
  exc.arg2 = (double)y;
  exc.type = UNDERFLOW;
  exc.name = (char *)"powf";
  if (sign == 1)
    exc.retval = zerof_with_flags(AMD_F_UNDERFLOW | AMD_F_INEXACT);
  else /* sign == -1 */
    exc.retval = -zerof_with_flags(AMD_F_UNDERFLOW | AMD_F_INEXACT);
  if (_LIB_VERSION == _POSIX_)
    __set_errno(ERANGE);
  else if (!matherr(&exc))
    __set_errno(ERANGE);
  return exc.retval;
}

/* Deal with errno for out-of-range arguments */
static inline float retval_errno_edom(float x, float y, int type)
{
  struct exception exc;
  exc.arg1 = (double)x;
  exc.arg2 = (double)y;
  exc.type = DOMAIN;
  exc.name = (char *)"powf";
  if (_LIB_VERSION == _SVID_)
    exc.retval = 0.0;
  else if (type == 1)
    exc.retval = infinityf_with_flags(AMD_F_DIVBYZERO);
  else if (type == 2)
    exc.retval = -infinityf_with_flags(AMD_F_DIVBYZERO);
  else /* type == 3 */
    exc.retval = nanf_with_flags(AMD_F_INVALID);
  if (_LIB_VERSION == _POSIX_)
    __set_errno (EDOM);
  if (!matherr(&exc))
    {
      if (_LIB_VERSION == _SVID_)
        (void)fputs("pow: DOMAIN error\n", stderr);
      __set_errno(EDOM);
    }
  return exc.retval;
}

float __powf(float x, float y)
{
  unsigned int ux, ax, uy, ay, mask;
  int yexp, inty, xpos, ypos, negateres;
  double dx, dy, dw, dlog2, dr;
  volatile int dummy;

  /* Largest float, stored as a double */
  const double large = 3.40282346638528859812e+38; /* 0x47efffffe0000000 */

  /* Smallest float, stored as a double */
  const double tiny = 1.40129846432481707092e-45; /* 0x36a0000000000000 */

  GET_BITS_SP32(x, ux);
  ax = ux & (~SIGNBIT_SP32);
  xpos = ax == ux;
  GET_BITS_SP32(y, uy);
  ay = uy & (~SIGNBIT_SP32);
  ypos = ay == uy;

  if (ux == 0x3f800000)
    {
      /* x = +1.0. Return +1.0 for all y, even NaN,
	 raising invalid only if y is a signalling NaN */
      if (y + 1.0F == 2.0F) dummy = 1;
      return 1.0F;
    }
  else if (ay == 0)
    {
      /* y is zero. Return 1.0, even if x is infinity or NaN,
         raising invalid only if x is a signalling NaN */
      if (x + 1.0F == 2.0F) dummy = 1;
      return 1.0F;
    }
  else if (((ax & EXPBITS_SP32) == EXPBITS_SP32) &&
           (ax & MANTBITS_SP32))
    /* x is NaN. Return NaN, with invalid exception if it's
       a signalling NaN. */
    return x + x;
  else if (((ay & EXPBITS_SP32) == EXPBITS_SP32) &&
           (ay & MANTBITS_SP32))
   /* y is NaN. Return NaN, with invalid exception if y
      is a signalling NaN. */
   return y + y;
  else if (uy == 0x3f800000)
    /* y is 1.0; return x */
    return x;
  else if ((ay & EXPBITS_SP32) > 0x4f000000)
    {
      /* y is infinite or so large that the result would 
         overflow or underflow. Flags should be raised 
         unless y is an exact infinity. */
      int yinf = (ay == EXPBITS_SP32);
      if (ypos)
        {
          /* y is +ve */
          if (ax == 0)
            /* abs(x) = 0.0. */
            return 0.0F;
          else if (ax < 0x3f800000)
	    {
	      /* abs(x) < 1.0 */
	      if (yinf)
                return 0.0F;
	      else
                return retval_errno_erange_underflow(x, y, 1);
	    }
          else if (ax == 0x3f800000)
            /* abs(x) = 1.0. */
            return 1.0F;
          else
	    {
	      /* abs(x) > 1.0 */
	      if (yinf)
                return infinityf_with_flags(0);
	      else
                return retval_errno_erange_overflow(x, y, 1);
	    }
        }
      else
        {
          /* y is -ve */
          if (ax == 0)
            /* abs(x) = 0.0. Return +infinity. */
            return retval_errno_edom(x, y, 1);
          else if (ax < 0x3f800000)
	    {
            /* abs(x) < 1.0; return +infinity. */
	      if (yinf)
                return infinityf_with_flags(0); 
	      else
                return retval_errno_erange_overflow(x, y, 1);
	    }
          else if (ax == 0x3f800000)
            /* abs(x) = 1.0. */
            return 1.0F;
          else
	    {
	      /* abs(x) > 1.0 */
	      if (yinf)
                return 0.0F;
	      else
                return retval_errno_erange_underflow(x, y, 1);
	    }
        }
    }

  /* See whether y is an integer.
     inty = 0 means not an integer.
     inty = 1 means odd integer.
     inty = 2 means even integer.
  */
  yexp = ((uy & EXPBITS_SP32) >> EXPSHIFTBITS_SP32) - EXPBIAS_SP32 + 1;
  if (yexp < 1)
    inty = 0;
  else if (yexp > 24)
    inty = 2;
  else /* 1 <= yexp <= 24 */
    {
      /* Mask out the bits of r that we don't want */
      mask = (1 << (24 - yexp)) - 1;
      if ((uy & mask) != 0)
        inty = 0;
      else if (((uy & ~mask) >> (24 - yexp)) & 0x00000001)
        inty = 1;
      else
        inty = 2;
    }

  if ((ax & EXPBITS_SP32) == EXPBITS_SP32)
    {
      /* x is infinity (NaN was already ruled out). */
      if (xpos)
        {
          /* x is +infinity */
          if (ypos)
            /* y > 0.0 */
            return x;
          else
            return 0.0F;
        }
      else
        {
          /* x is -infinity */
          if (inty == 1)
            {
              /* y is an odd integer */
              if (ypos)
                /* Result is -infinity */
                return x;
              else
                return -0.0F;
            }
          else
            {
              if (ypos)
                /* Result is +infinity */
                return -x;
              else
                return 0.0F;
            }
        }
    }
  else if (ax == 0)
    {
      /* x is zero */
      if (xpos)
        {
          /* x is +0.0 */
          if (ypos)
            /* y is positive; return +0.0 for all cases */
            return x;
          else
            /* y is negative; return +infinity with div-by-zero
               for all cases */
            return retval_errno_edom(x, y, 1);
        }
      else
        {
          /* x is -0.0 */
          if (ypos)
            {
              /* y is positive */
              if (inty == 1)
                /* -0.0 raised to a positive odd integer returns -0.0 */
                return x;
              else
                /* Return +0.0 */
                return -x;
            }
          else
            {
              /* y is negative */
              if (inty == 1)
                /* -0.0 raised to a negative odd integer returns -infinity
                   with div-by-zero */
                return retval_errno_edom(x, y, 2);
              else
                /* Return +infinity with div-by-zero */
                return retval_errno_edom(x, y, 1);
            }
        }
    }

  negateres = 0;
  if (!xpos)
    {
      /* x is negative */
      if (inty)
        {
          /* It's OK because y is an integer. */
          ux = ax;
          PUT_BITS_SP32(ux, x); /* x = abs(x) */
          /* If y is odd, the result will be negative */
          negateres = (inty == 1);
        }
      else
        /* y is not an integer. Return a NaN. */
        return retval_errno_edom(x, y, 3);
    }

  if (ay < 0x2e800000)   /* abs(y) < 2^(-34) */
    {
      /* y is close enough to zero for the result to be 1.0
         no matter what the size of x */
      return 1.0F + y;
    }

  /* Simply use double precision for computation of log2(x),
     y*log2(x) and exp2(y*log2(x)) */
  dx = x;
  dy = y;
  dlog2 = log2(dx);
  dw = y * dlog2;
  dr = exp2(dw);

  /* If dr overflowed or underflowed we need to deal with errno */
  if (dr > large)
    {
      /* Double dr has overflowed range of float. */
      if (negateres)
        return retval_errno_erange_overflow(x, y, -1);
      else
        return retval_errno_erange_overflow(x, y, 1);
    }
  else if (dr < tiny)
    {
      /* Double dr has underflowed range of float. */
      if (negateres)
        return retval_errno_erange_underflow(x, y, -1);
      else
        return retval_errno_erange_underflow(x, y, 1);
    }
  else
    {
      if (negateres)
        return -dr;
      else
        return dr;
    }
}

weak_alias (__powf, powf)
weak_alias (__powf, __ieee754_powf)
